"""Fields module."""

import datetime

import six

from . import errors


class Field(property):
    """Base field."""

    _converter = lambda _, value: value
    """Convert raw input value of the field."""

    def __init__(self, default=None):
        """Initializer."""
        super(Field, self).__init__(self.get_value, self.set_value)
        self.name = None
        self.storage_name = None

        self.model_cls = None

        self.default = default

    def bind_name(self, name):
        """Bind field to its name in model class."""
        if self.name:
            raise errors.Error('Already bound "{0}" with name "{1}" could not '
                               'be rebound'.format(self, self.name))
        self.name = name
        self.storage_name = ''.join(('_', self.name))
        return self

    def bind_model_cls(self, model_cls):
        """Bind field to model class."""
        if self.model_cls:
            raise errors.Error('"{0}" has been already bound to "{1}" and '
                               'could not be rebound to "{2}"'.format(
                                   self, self.model_cls, model_cls))
        self.model_cls = model_cls
        return self

    def init_model(self, model, value):
        """Init model with field."""
        if not value:
            value = self.default() if callable(self.default) else self.default
        setattr(model, self.storage_name, value)

    def get_value(self, model):
        """Return field's value."""
        return getattr(model, self.storage_name)

    def set_value(self, model, value):
        """Set field's value."""
        if value is not None:
            value = self._converter(value)
        setattr(model, self.storage_name, value)


class Bool(Field):
    """Bool field."""

    _converter = lambda _, value: bool(value)
    """Convert raw input value of the field."""


class Int(Field):
    """Int field."""

    _converter = lambda _, value: int(value)
    """Convert raw input value of the field."""


class Float(Field):
    """Float field."""

    _converter = lambda _, value: float(value)
    """Convert raw input value of the field."""


class String(Field):
    """String field."""

    _converter = lambda _, value: str(value)
    """Convert raw input value of the field."""


class Binary(Field):
    """Binary field."""

    _converter = lambda _, value: six.binary_type(value)
    """Convert raw input value of the field."""


class Date(Field):
    """Date field."""

    def _converter(self, value):
        """Convert raw input value of the field."""
        if not isinstance(value, datetime.date):
            raise TypeError('{0} is not valid date'.format(value))
        return value


class DateTime(Field):
    """Date and time field."""

    def _converter(self, value):
        """Convert raw input value of the field."""
        if not isinstance(value, datetime.datetime):
            raise TypeError('{0} is not valid date and time')
        return value


class Model(Field):
    """Model relation field."""

    def __init__(self, related_model_cls, default=None):
        """Initializer."""
        super(Model, self).__init__(default=default)

        self.related_model_cls = related_model_cls

    def _converter(self, value):
        """Convert raw input value of the field."""
        if not isinstance(value, self.related_model_cls):
            raise TypeError('{0} is not valid model instance, instance of '
                            '{1} required'.format(value,
                                                  self.related_model_cls))
        return value


class Collection(Field):
    """Models collection relation field."""

    def __init__(self, related_model_cls, default=None):
        """Initializer."""
        super(Collection, self).__init__(default=default)
        self.related_model_cls = related_model_cls

    def _converter(self, value):
        """Convert raw input value of the field."""
        if type(value) is not self.related_model_cls.Collection:
            value = self.related_model_cls.Collection(value)
        return value



"""Models module."""

from __future__ import absolute_import

import collections as std_collections
import six

from . import fields
from . import collections
from . import errors


class DomainModelMetaClass(type):
    """Domain model meta class."""

    def __new__(mcs, class_name, bases, attributes):
        """Domain model class factory."""
        model_fields = mcs.parse_fields(attributes)

        if attributes.get('__slots_optimization__', True):
            attributes['__slots__'] = mcs.prepare_model_slots(model_fields)

        cls = type.__new__(mcs, class_name, bases, attributes)

        cls.__fields__ = mcs.bind_fields_to_model_cls(cls, model_fields)
        cls.__unique_key__ = mcs.prepare_fields_attribute(
            attribute_name='__unique_key__', attributes=attributes,
            class_name=class_name)
        cls.__view_key__ = mcs.prepare_fields_attribute(
            attribute_name='__view_key__', attributes=attributes,
            class_name=class_name)

        mcs.bind_collection_to_model_cls(cls)

        return cls

    @staticmethod
    def parse_fields(attributes):
        """Parse model fields."""
        return tuple(field.bind_name(name)
                     for name, field in six.iteritems(attributes)
                     if isinstance(field, fields.Field))

    @staticmethod
    def prepare_model_slots(model_fields):
        """Return tuple of model field slots."""
        return tuple(field.storage_name for field in model_fields)

    @staticmethod
    def prepare_fields_attribute(attribute_name, attributes, class_name):
        """Prepare model fields attribute."""
        attribute = attributes.get(attribute_name)
        if not attribute:
            attribute = tuple()
        elif isinstance(attribute, std_collections.Iterable):
            attribute = tuple(attribute)
        else:
            raise errors.Error('{0}.{1} is supposed to be a list of {2}, '
                               'instead {3} given', class_name, attribute_name,
                               fields.Field, attribute)
        return attribute

    @staticmethod
    def bind_fields_to_model_cls(cls, model_fields):
        """Bind fields to model class."""
        return tuple(field.bind_model_cls(cls)
                     for field in model_fields)

    @staticmethod
    def bind_collection_to_model_cls(cls):
        """Bind collection to model's class.

        If collection was not specialized in process of model's declaration,
        subclass of collection will be created.
        """
        cls.Collection = type('{0}.Collection'.format(cls.__name__),
                              (cls.Collection,),
                              {'value_type': cls})
        cls.Collection.__module__ = cls.__module__


@six.python_2_unicode_compatible
@six.add_metaclass(DomainModelMetaClass)
class DomainModel(object):
    """Base domain model.

    .. py:attribute:: Collection

        Model's collection class.

        :type: collections.Collection

    .. py:attribute:: __fields__

        Tuple of all model fields.

        :type: tuple[fields.Field]

    .. py:attribute:: __unique_key__

        Tuple of model fields that represents unique key.

        :type: tuple[fields.Field]

    .. py:attribute:: __view_key__

        Tuple of model fields that represents view key.

        :type: tuple[fields.Field]
    """

    Collection = collections.Collection

    __fields__ = tuple()
    __view_key__ = tuple()
    __unique_key__ = tuple()
    __slots_optimization__ = True

    def __init__(self, **kwargs):
        """Initializer."""
        for field in self.__class__.__fields__:
            field.init_model(self, kwargs.get(field.name))
        super(DomainModel, self).__init__()

    def __eq__(self, other):
        """Make equality comparation based on unique key.

        If unique key is not defined, standard object's equality comparation
        will be used.
        """
        if self is other:
            return True
        if not isinstance(other, self.__class__):
            return False
        if not self.__class__.__unique_key__:
            return NotImplemented
        for field in self.__class__.__unique_key__:
            if field.get_value(self) != field.get_value(other):
                return False
        return True

    def __ne__(self, other):
        """Make non-equality comparation based on unique key.

        If unique key is not defined, standard object's not equality
        comparation will be used.
        """
        if isinstance(other, self.__class__):
            return not self == other
        return NotImplemented

    def __hash__(self):
        """Calculate and return model hash based on unique key.

        If unique key is not defined, standard object's hash calculation will
        be used.
        """
        if self.__class__.__unique_key__:
            return hash(tuple(field.get_value(self)
                              for field in self.__class__.__unique_key__))
        return super(DomainModel, self).__hash__()

    def __repr__(self):
        """Return Pythonic representation of domain model."""
        return '{module}.{cls}({fields_values})'.format(
            module=self.__class__.__module__, cls=self.__class__.__name__,
            fields_values=', '.join('='.join((field.name,
                                              repr(field.get_value(self))))
                                    for field in self.__class__.__fields__))

    def __str__(self):
        """Return string representation of domain model."""
        if not self.__class__.__view_key__:
            return self.__repr__()
        return '{module}.{cls}({fields_values})'.format(
            module=self.__class__.__module__, cls=self.__class__.__name__,
            fields_values=', '.join('='.join((field.name,
                                              str(field.get_value(self))))
                                    for field in self.__class__.__view_key__))

    @property
    def __data__(self):
        """Read only dictionary of model fields/values."""
        return dict((field.name, field.get_value(self))
                    for field in self.__class__.__fields__)



"""Collections module."""

import six


class Collection(list):
    """Collection."""

    value_type = object
    """Type of values that collection could contain."""

    def __init__(self, iterable=None, type_check=True):
        """Initializer."""
        if not iterable:
            iterable = tuple()

        if type_check:
            iterable = self._ensure_iterable_is_valid(iterable)

        super(Collection, self).__init__(iterable)

    def append(self, value):
        """Add an item to the end of the list."""
        return super(Collection, self).append(
            self._ensure_value_is_valid(value))

    def extend(self, iterable):
        """Extend the list by appending all the items in the given list."""
        return super(Collection, self).extend(
            self._ensure_iterable_is_valid(iterable))

    def insert(self, index, value):
        """Insert an item at a given position."""
        return super(Collection, self).insert(
            index, self._ensure_value_is_valid(value))

    def __setitem__(self, index, value):
        """Set an item at a given position."""
        if isinstance(index, slice):
            return super(Collection, self).__setitem__(index,
                                                       self.__class__(value))
        else:
            return super(Collection, self).__setitem__(
                index, self._ensure_value_is_valid(value))

    def __getitem__(self, index):
        """Return value by index or slice of values if index is slice."""
        value = super(Collection, self).__getitem__(index)
        if isinstance(index, slice):
            return self.__class__(value, type_check=False)
        return value

    if six.PY2:  # pragma: nocover
        def __getslice__(self, start, stop):
            """Return slice of values."""
            return self.__class__(super(Collection, self).__getslice__(start,
                                                                       stop),
                                  type_check=False)

        def __setslice__(self, start, stop, iterable):
            """Set slice of values."""
            super(Collection, self).__setslice__(start, stop,
                                                 self.__class__(iterable))

    def _ensure_iterable_is_valid(self, iterable):
        """Ensure that iterable values are a valid collection's values."""
        for value in iterable:
            self._ensure_value_is_valid(value)
        return iterable

    def _ensure_value_is_valid(self, value):
        """Ensure that value is a valid collection's value."""
        if not isinstance(value, self.__class__.value_type):
            raise TypeError('{0} is not valid collection value, instance '
                            'of {1} required'.format(
                                value, self.__class__.value_type))
        return value



"""`Domain models` setup script."""

import os
import re

from setuptools import setup
from setuptools import Command

SHORT_DESCRIPTION = 'Domain models framework for Python projects'


# Getting description:
with open('README.rst') as readme_file:
    description = readme_file.read()

    # Removing duplicated short description.
    description = description.replace(SHORT_DESCRIPTION, '')

# Getting requirements:
with open('requirements.txt') as version:
    requirements = version.readlines()

# Getting version:
with open('domain_models/__init__.py') as init_file:
    version = re.search('VERSION = \'(.*?)\'', init_file.read()).group(1)


class PublishCommand(Command):
    """Setuptools `publish` command."""

    description = "Publish current distribution to PyPi and create tag"
    user_options = []

    def initialize_options(self):
        """Init options."""

    def finalize_options(self):
        """Finalize options."""

    def run(self):
        """Command execution."""
        self.run_command('sdist')
        self.run_command('upload')
        os.system('git tag -a {0} -m \'version {0}\''.format(version))
        os.system('git push --tags')


setup(name='domain_models',
      version=version,
      description=SHORT_DESCRIPTION,
      long_description=SHORT_DESCRIPTION,
      author='ETS Labs',
      author_email='rmogilatov@gmail.com',
      maintainer='ETS Labs',
      maintainer_email='rmogilatov@gmail.com',
      url='https://github.com/ets-labs/domain_models',
      license='BSD New',
      packages=['domain_models'],
      zip_safe=True,
      install_requires=requirements,
      cmdclass={
          'publish': PublishCommand,
      },
      keywords=[
          'Domain models',
          'Domain modelling',
          'Domain driven design',
          'Domain driven development',
          'DDD',
          'Models',
      ],
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Topic :: Software Development',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ])
